function init(tilemapid,roomsize_,tilesize_)
	map = TileMap.new(tilemapid)
	roomsize = roomsize_
	tilesize = tilesize_
	features = {}
	visualfeatures = {}
	featureindex = {}
	objectdata = {}
	units = {}
	tiledunits = {}
	codeunits = {}
	unitlists = {}
	objectlist = {}
	undobuffer = {}
	animunits = {}
	unitmap = {}
	deleted = {}
	currid = 0
	currdir = 4
	updatecode = 1
	doundo = true
	updateundo = true
	ruledebug = false
	maprotation = 0
	ruleids = {}
	updatelist = {}
end

function addunit(id)
	local unitid = #units + 1
	
	units[unitid] = {}
	units[unitid] = mmf.newObject(id)
	
	local unit = units[unitid]
	
	if (unit.values[ID] == -1) then
		unit.values[ID] = newid()
	end

	if (unit.values[XPOS] > 0) and (unit.values[YPOS] > 0) then
		addunitmap(id,unit.values[XPOS],unit.values[YPOS])
	end
	
	if (unit.values[TILING] == 1) then
		table.insert(tiledunits, unit.fixed)
	end
	
	if (unit.values[TILING] > 1) then
		table.insert(animunits, unit.fixed)
	end
	
	local name = unit.strings[UNITNAME]
	local name_ = unit.strings[NAME]
	
	if (unitlists[name] == nil) then
		unitlists[name] = {}
	end
	
	if (unit.strings[UNITTYPE] ~= "text") or ((unit.strings[UNITTYPE] == "text") and (unit.values[TYPE] == 0)) then
		objectlist[name_] = 1
	end
	
	table.insert(unitlists[name], unit.fixed)
	
	if (unit.strings[UNITTYPE] == "text") then
		table.insert(codeunits, unit.fixed)
		updatecode = 1
		
		if (unit.values[TYPE] == 0) then
			local matname = string.sub(unit.strings[UNITNAME], 6)
			if (unitlists[matname] == nil) then
				unitlists[matname] = {}
			end
		end
	end
	
	setcolour(unit.fixed)
end

function clearunits()
	units = {}
	tiledunits = {}
	codeunits = {}
	animunits = {}
	unitlists = {}
	undobuffer = {}
	unitmap = {}
	prevunitmap = {}
	currid = 0
	updateundo = true
	newundo()
	currdir = 4
	ruleids = {}
	objectlist = {}
	updatelist = {}
end

function smallclear()
	objectdata = {}
	deleted = {}
	updatelist = {}
end

function clear()
	features = {}
	featureindex = {}
	visualfeatures = {}
	objectdata = {}
	deleted = {}
	updatecode = 1
	updateundo = false
	ruleids = {}
	updatelist = {}
end

function command(key)
	--local special = featureindex[key]
	local keyid = -1
	if (keys[key] ~= nil) then
		keyid = keys[key]
	else
		print("no such key")
		return
	end

	if (special == nil) then
		if (keyid <= 4) then
			local drs = ndirs[keyid+1]
			local ox = drs[1]
			local oy = drs[2]
			local dir = keyid
			
			if (dir == 4) then
				dir = nil
			end
			
			MF_update()
			movecommand(ox,oy,dir)
		end
		
		if (keyid == 5) then
			MF_restart()
		end
	else
		MF_update()
	end
end

function findfeature(rule1,rule2,rule3)
	local options = {}
	local result = {}
	local rule = ""
	
	if (rule1 ~= nil) then
		rule = rule1 .. " "
	end
	
	if (rule2 ~= nil) then
		rule = rule .. rule2 .. " "
	end
	
	if (rule3 ~= nil) then
		rule = rule .. rule3
	end
	
	if (featureindex[rule1] ~= nil) then
		for i,rules in ipairs(featureindex[rule1]) do
			local rule = rules[1]
			local conds = rules[2]
			
			if (conds[1] ~= "never") then
				if (rule[1] == rule1) and (rule[2] == rule2) then
					local baserule = {rule[1],rule[2],rule[3]}
					table.insert(options, {baserule,conds})
				end
			end
		end
	end
	
	if (featureindex[rule3] ~= nil) then
		for i,rules in ipairs(featureindex[rule3]) do
			local rule = rules[1]
			local conds = rules[2]
			
			if (conds[1] ~= "never") then
				if (rule[3] == rule3) and (rule[2] == rule2) then
					local baserule = {rule[1],rule[2],rule[3]}
					table.insert(options, {baserule,conds})
				end
			end
		end
	end
	
	if (rule1 == nil) and (rule3 == nil) and (rule2 ~= nil) then
		if (featureindex[rule2] ~= nil) then 
			for i,rules in ipairs(featureindex[rule2]) do
				local usable = false
				local rule = rules[1]
				local conds = rules[2]

				if (conds[1] ~= "never") then
					for a,mat in pairs(objectlist) do
						if (a == rule[3]) then
							usable = true
						end
					end
					
					if (rule[2] == rule2) and usable then
						local baserule = {rule[1],rule[2],rule[3]}
						table.insert(options, {baserule,conds})
					end
				end
			end
		end
	end
	
	for i,rules in ipairs(options) do
		local words = {}
		local limit = 0
		local baserule = rules[1]
		
		for a,b in ipairs(baserule) do
			table.insert(words, b)
			limit = limit + 1
			
			if (limit == 4) then
				break
			end
		end
		
		if (#words >= 3) then
			local one = words[3]
			local two = words[2] .. " " .. words[3]
			local three = words[1] .. " " .. words[2] .. " " .. words[3]

			if (one == rule) or (two == rule) or (three == rule) or ((rule2 == words[2]) and (rule1 == nil) and (rule3 == nil)) then				
				table.insert(result, {baserule[1], rules[2]})
			end
		end
	end
	
	if (#result > 0) then
		return result
	else
		return nil
	end
end

function createall(matdata,x_,y_,id_)
	local all = {}
	local empty = false
	
	if (x_ == nil) and (y_ == nil) and (id_ == nil) then
		if (matdata[1] ~= "empty") and (matdata[1] ~= "level") and (matdata[1] ~= "group") then
			all = findall(matdata)
		else
			all = findempty()
			empty = true
		end
	end
	local test = {}
	
	if (x_ ~= nil) and (y_ ~= nil) and (id_ ~= nil) then
		local check = findtype(matdata,x_,y_,id_)
		
		if (#check > 0) then
			for i,v in ipairs(check) do
				if (v ~= 0) then
					table.insert(test, v)
				end
			end
		end
	end
	
	if (#all > 0) then
		for i,v in ipairs(all) do
			table.insert(test, v)
		end
	end
	
	if (#test > 0) then
		for i,v in ipairs(test) do
			if (empty == false) then
				local vunit = mmf.newObject(v)
				local x,y,dir = vunit.values[XPOS],vunit.values[YPOS],vunit.values[DIR],vunit.values[MOVED]
						
				for b,unit in pairs(objectlist) do
					if (b ~= "empty") and (b ~= "all") and (b ~= "level") and (b ~= "group") and (b ~= matdata[1]) then
						local mat = findtype({b},x,y,v)
						local tmat = findtext(x,y)
						
						if ((#mat == 0) and (#tmat == 0)) or ((#mat == 0) and (matdata[1] == "text")) then
							create(b,x,y,dir)
							
							if (matdata[1] == "text") and (#tmat > 0) then
								for c,d in ipairs(tmat) do
									local tunit = mmf.newObject(d)
									
									if (tunit.strings[UNITNAME] == "text_" .. b) then
										delete(d,0)
									end
								end
							end
						end
					end
				end
			else
				local x = v % roomsize
				local y = math.floor(v / roomsize)
				local dir = 4
				
				for b,mat in pairs(objectlist) do
					if (b ~= "empty") and (b ~= "all") and (b ~= "level") and (b ~= "group") then
						local mat = findtype({b},x,y,0)
						local tmat = findtext(x,y)
						
						if ((#mat == 0) and (tmat == 0)) or ((#mat == 0) and (matdata[1] == "text")) then
							create(b,x,y,dir)
							
							if (matdata[1] == "text") and (tmat ~= 0) then
								delete(tmat,0)
							end
						end
					end
				end
			end
		end
	end
end

function setunitmap()
	unitmap = {}
	
	for i,unit in ipairs(units) do
		local tileid = unit.values[XPOS] + unit.values[YPOS] * roomsize
		
		--print(tostring(unit.values[XPOS]) .. ", " .. tostring(unit.values[YPOS]) .. ", " .. unit.strings[UNITNAME])
		
		if (unitmap[tileid] == nil) then
			unitmap[tileid] = {}
		end

		table.insert(unitmap[tileid], unit.fixed)
	end
end

function setundo(this)
	if (this ~= nil) then
		if (this == 1) then
			updateundo = true
		elseif (this == 0) then
			updateundo = false
		end
	else
		print("undo is nil!")
		updateundo = true
	end
end

function victory()
	MF_win()
end

function setcolour(unitid,value_)
	local unit = mmf.newObject(unitid)
	
	local name = unit.strings[UNITNAME]
	local unitinfo = tileslist[name]
	
	if (value_ == nil) then
		if (unitinfo == nil) or (unitinfo.colour == nil) then
			MF_defaultcolour(unitid)
		elseif (unitinfo ~= nil) and (unitinfo.colour ~= nil) then
			local colour = unitinfo.colour
			MF_setcolour(unitid,colour[1],colour[2])
		end
	else
		if (unitinfo == nil) or (unitinfo[value_] == nil) then
			MF_defaultcolour(unitid)
		elseif (unitinfo ~= nil) and (unitinfo[value_] ~= nil) then
			local colour = unitinfo[value_]
			MF_setcolour(unitid,colour[1],colour[2])
		end
	end
end

function getcolour(unitid,value_)
	local unit = mmf.newObject(unitid)
	
	local name = unit.strings[UNITNAME]
	local unitinfo = tileslist[name]
	
	local defaultcolour = colours.default
	
	local value = "colour"
	
	if (value_ ~= nil) then
		value = value_
	end
	
	if (unitinfo == nil) then
		return defaultcolour[1],defaultcolour[2]
	else
		if (unitinfo[value] == nil) then
			if (unitinfo.colour == nil) then
				return defaultcolour[1],defaultcolour[2]
			else
				local colour = unitinfo.colour
				return colour[1],colour[2]
			end
		else
			local colour = unitinfo[value]
			return colour[1],colour[2]
		end
	end

	return defaultcolour[1],defaultcolour[2]
end

function getuicolour(which,subwhich)
	local bcolour = {}
	
	if (subwhich == nil) or (subwhich == "") then
		bcolour = colours[which]
	else
		bcolour = colours[which][subwhich]
	end
	
	if (bcolour == nil) then
		bcolour = colours.default
	end
	
	return bcolour[1],bcolour[2]
end

function updatecolours()
	MF_setbackcolour()
	
	for i,unit in ipairs(units) do
		if (unit.strings[UNITNAME] ~= "level") then
			setcolour(unit.fixed)
		end
	end
	
	updatecode = 1
	code()
end

function poscorrect(unitid,rotation,zoom)
	local unit = mmf.newObject(unitid)
	
	local midpoint = roomsize * tilesize * 0.5
	local midtile = math.floor(roomsize * 0.5) - 0.5
	
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local dx = x - midtile
	local dy = y - midtile
	
	local dir = 0 - math.atan2(dy,dx) + math.rad(rotation)
	local dist = math.sqrt((dy)^2 + (dx)^2)
	
	local newx = midpoint + math.cos(dir) * dist * zoom * tilesize
	local newy = midpoint - math.sin(dir) * dist * zoom * tilesize
	
	if (unit.values[FLOAT] == 0) then
		unit.x = newx
		unit.y = newy
	end
end

function fullstep(updatestyle)
	if (1 == 0) then
		local style = 1
		
		if (updatestyle ~= nil) then
			style = updatestyle
		end
		
		if (style == 1) then
			MF_removedebug()
		end
		
		doupdate()
		smallclear()
		setunitmap()
		
		code()
		
		if (style == 1) then
			moveblock()
			fallblock()
			code()
			keyblock()
			block()
			code()
			
			-- SOUND THINGS HERE
			
			conversion()
			fallblock()
			
			doupdate()
			smallclear()
			setunitmap()
			MF_checked()
			
			code()
		end
		
		musicstate()
		setunitmap()
		MF_checked()
		MF_flag17()
		
		if (style == 1) then
			MF_setmovement()
			newundo()
		end
		
		animate()
	end
end