function rotate(dir)
	local rot = {2,3,0,1}
	
	return rot[dir+1]	
end

function loadtile(x,y)
	local key = tostring(x) .. "," .. tostring(y)
	
	if (tiles[key] ~= nil) then
		return tiles[key]
	else
		return "error"
	end
end

function isthis(lookup,rule)
	for i,rules in ipairs(lookup) do
		local baserule = rules[1]
		
		if (baserule[3] == rule) then
			return true
		end
	end
	
	return false
end

function xthis(lookup,name,rule)
	local result = {}
	
	for i,rules in ipairs(lookup) do
		local baserule = rules[1]
		
		if (baserule[1] == name) and (baserule[2] == rule) then
			table.insert(result, baserule[3])
		end
	end
	
	return result
end

function dynamictile(unitid,x,y,name)
	local ox,oy = 0,0
	local result = 0
	local exclude = 0
	local layer = map[0]
	
	for i=1,4 do
		local v = ndirs[i]
		ox = v[1]
		oy = v[2]
		
		local tiles = findtype({name},x+ox,y+oy,unitid)
		local maptile = 0
		
		if (x+ox >= 0) and (y+oy >= 0) and (x+ox < roomsize) and (y+oy < roomsize) then
			maptile = layer:get_x(x+ox,y+oy)
		end
		
		if (#tiles > 0) or (maptile ~= 255) then
			result = result + 2 ^ (i - 1)
		end
	end
	
	return result
end

function dynamictiling()
	for i,unitid in ipairs(tiledunits) do
		local unit = mmf.newObject(unitid)
		unit.direction = dynamictile(unitid,unit.values[XPOS],unit.values[YPOS],unit.strings[UNITNAME])
	end
end

function dynamic(id)
	local unit = mmf.newObject(id)
	
	if (unit.values[TILING] == 1) then
		local x,y = unit.values[XPOS],unit.values[YPOS]
		local ox,oy = 0,0
		local name = unit.strings[UNITNAME]
		
		unit.direction = dynamictile(unit.fixed,x,y,name)
		
		for i=1,4 do
			local v = ndirs[i]
			ox = v[1]
			oy = v[2]
			
			local tiles = findtype({name},x+ox,y+oy,unit.fixed)
			
			if (#tiles > 0) then
				for a,b in ipairs(tiles) do
					local tile = mmf.newObject(b)

					tile.direction = dynamictile(b,x+ox,y+oy,name)
				end
			end
		end
	end
end

function dynamicat(x,y)
	local ox,oy = 0,0
	
	for i=1,4 do
		local v = ndirs[i]
		ox = v[1]
		oy = v[2]
		
		local tiles = findallhere(x+ox,y+oy)
		
		if (#tiles > 0) then
			for a,b in ipairs(tiles) do
				local tile = mmf.newObject(b)

				if (tile.values[TILING] == 1) then
					tile.direction = dynamictile(b,x+ox,y+oy,tile.strings[UNITNAME])
				end
			end
		end
	end
end

function findall(name_)
	local result = {}
	local name = name_[1]
	
	for i,unit in ipairs(units) do
		local unitname = unit.strings[UNITNAME]
		
		if (unit.strings[UNITTYPE] == "text") then
			unitname = "text"
		end
		
		if (unitname == name) then
			if testcond(name_[2],unit.fixed) then
				table.insert(result, unit.fixed)
			end
		end
	end
	
	return result
end

function delunit(unitid)
	local unit = mmf.newObject(unitid)
	local name = unit.strings[UNITNAME]
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local unitlist = unitlists[name]
	local unittype = unit.strings[UNITTYPE]
	
	if (unittype == "text") then
		updatecode = 1
	end
	
	for i,v in ipairs(unitlist) do
		if (v == unitid) then
			v = {}
			table.remove(unitlist, i)
		end
	end
	
	-- TÄMÄ EI EHKÄ TOIMI
	local tileid = x + y * roomsize
	for i,v in ipairs(unitmap[tileid]) do
		if (v == unitid) then
			v = {}
			table.remove(unitmap[tileid], i)
		end
	end
	
	if (unit.strings[UNITTYPE] == "text") then
		for i,v in ipairs(codeunits) do
			if (v == unitid) then
				v = {}
				table.remove(codeunits, i)
			end
		end
	end
	
	if (unit.values[TILING] > 1) then
		for i,v in ipairs(animunits) do
			if (v == unitid) then
				v = {}
				table.remove(animunits, i)
			end
		end
	end
	
	if (unit.values[TILING] == 1) then
		for i,v in ipairs(tiledunits) do
			if (v == unitid) then
				v = {}
				table.remove(tiledunits, i)
			end
		end
	end
	
	for i,v in ipairs(units) do
		if (v.fixed == unitid) then
			v = {}
			table.remove(units, i)
		end
	end
end

function findtype(typedata,x,y,unitid_)
	local result = {}
	local unitid = 0
	local tileid = x + y * roomsize
	local name = typedata[1]
	local conds = typedata[2]
	
	if (unitid_ ~= nil) then
		unitid = unitid_
	end

	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			if (v ~= unitid) then
				local unit = mmf.newObject(v)
				
				if (unit.strings[UNITNAME] == name) or ((unit.strings[UNITTYPE] == "text") and (name == "text")) then
					if testcond(conds,v) then
						table.insert(result, v)
					end
				end
			end
		end
	end
	
	return result
end

function findobstacle(x,y)
	local layer = map[0]
	local tile = layer:get_x(x,y)
	local result = {}
	local tileid = x + y * roomsize
	
	if (tile ~= 255) then
		table.insert(result, -1)
	end
	
	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			local unit = mmf.newObject(v)
			
			if (unit.flags[15] == false) then
				table.insert(result, v)
			else
				print("Prevunitmap has a removed object: " .. unit.strings[UNITNAME] .. ", " .. tostring(v))
			end
		end
	end
	
	return result
end

function update(unitid,x,y,dir_)
	updateundo = true
	local unit = mmf.newObject(unitid)

	local unitname = unit.strings[UNITNAME]
	local dir,olddir = unit.values[DIR],unit.values[DIR]
	local tiling = unit.values[TILING]
	local unittype = unit.strings[UNITTYPE]
	local oldx,oldy = unit.values[XPOS],unit.values[YPOS]
	
	if (dir_ ~= nil) then
		dir = dir_
	end
	
	addundo({"update",unitname,oldx,oldy,olddir,x,y,dir,unit.values[ID]})
	
	unit.values[XPOS] = x
	unit.values[YPOS] = y
	unit.values[DIR] = dir
	unit.values[MOVED] = 1
	unit.values[POSITIONING] = 0

	updateunitmap(unitid,oldx,oldy,x,y)
	
	if (tiling == 1) then
		dynamic(unitid)
		dynamicat(oldx,oldy)
	end
	
	if (unittype == "text") then
		updatecode = 1
	end
end

function updatedir(unitid,dir)
	updateundo = true
	local unit = mmf.newObject(unitid)
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local unitname = unit.strings[UNITNAME]
	local unittype = unit.strings[UNITTYPE]
	local olddir = unit.values[DIR]
	
	addundo({"update",unitname,x,y,olddir,x,y,dir,unit.values[ID]})
	unit.values[DIR] = dir
	
	if (unittype == "text") then
		updatecode = 1
	end
end

function findtext(x,y)
	local result = {}
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] ~= nil) then
		for i,v in ipairs(unitmap[tileid]) do
			local unit = mmf.newObject(v)
			
			if (unit.strings[UNITTYPE] == "text") then
				table.insert(result, v)
			end
		end
	end
	
	return result
end

function findallhere(x,y,exclude_)
	local result = {}
	
	local exclude = 0
	
	if (exclude_ ~= nil) then
		exclude = exclude_
	end
	
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] ~= nil) then
		for i,unitid in ipairs(unitmap[tileid]) do
			if (unitid ~= exclude) then
				local unit = mmf.newObject(unitid)
				if (unit.values[XPOS] == x) and (unit.values[YPOS] == y) then
					table.insert(result, unitid)
				end
			end
		end
	end
	
	return result
end

function findempty()
	local result = {}
	local array = {}
	local layer = map[0]
	
	for i,unit in ipairs(units) do
		local x,y = unit.values[XPOS],unit.values[YPOS]
		local arrayid = x + y * roomsize
		
		if (array[arrayid] == nil) then
			array[arrayid] = {}
			table.insert(array[arrayid], unit.fixed)
		end
	end
	
	for i=0,roomsize-1 do
		for j=0,roomsize-1 do
			local empty = 1
			local tile = layer:get_x(i,j)
			local arrayid = i + j * roomsize
			
			if (tile ~= 255) or (array[arrayid] ~= nil) then
				empty = 0
			end
			
			if (empty == 1) then
				table.insert(result, i + j * roomsize)
			end
		end
	end	
	
	return result
end

function delete(unitid,shake,sound_)
	local sound = 1
	
	if (sound_ ~= nil) then
		sound = sound_
	end
	
	if (deleted[unitid] == nil) then
		local unit = mmf.newObject(unitid)
		
		local x,y,dir = unit.values[XPOS],unit.values[YPOS],unit.values[DIR]
		local unitname = unit.strings[UNITNAME]
		local insidename = unit.strings[UNITNAME]
		
		if (unit.strings[UNITTYPE] == "text") then
			insidename = "text"
		end
		
		inside(unitname,x,y,dir,unitid)
		
		addundo({"remove",unitname,x,y,dir,unit.values[ID]})
		unit = {}
		delunit(unitid)
		MF_remove(unitid,shake,sound)
		
		dynamicat(x,y)
		
		deleted[unitid] = 1
	else
		print("already deleted")
	end
end

function findwalls(x,y)
	local result = {}
	local stop = findfeature(nil,"is","stop")
	local push = findfeature(nil,"is","push")
	local pull = findfeature(nil,"is","pull")
	local layer = map[0]
	
	if (stop ~= nil) then
		for i,v in ipairs(stop) do
			local stops = findtype(v,x,y,0)
			
			if (#stops > 0) then
				for a,b in ipairs(stops) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (push ~= nil) then
		for i,v in ipairs(push) do
			local pushes = findtype(v,x,y,0)
			
			if (#pushes > 0) then
				for a,b in ipairs(pushes) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (pull ~= nil) then
		for i,v in ipairs(pull) do
			local pulls = findtype(v,x,y,0)
			
			if (#pulls > 0) then
				for a,b in ipairs(pulls) do
					table.insert(result, b)
				end
			end
		end
	end
	
	if (layer:get_x(x,y) ~= 255) then
		table.insert(result, 1)
	end
	
	return result
end

function listrules()
	print("Units: " .. tostring(#units) .. "; Rules: " .. tostring(#features) .. "; Code: " .. tostring(#codeunits) .. "; Map: " .. tostring(mapcells()))
	print("Listing the rules:")
	
	for i,rules in ipairs(features) do
		local text = ""
		local rule = rules[1]
		
		for a,b in ipairs(rule) do
			text = text .. b .. " "
		end
		
		local conds = rules[2]
		if (#conds > 0) then
			for a,cond in ipairs(conds) do
				text = text .. cond[1] .. " "
				
				if (cond[2] ~= nil) then
					if (#cond[2] > 0) then
						for c,d in ipairs(cond[2]) do
							text = text .. d .. " "
						end
					end
				end
			end
		end
		
		print(text)
	end
	
	updatecode = 1
	code()
end

function writerules()
	MF_write("rules:",0)
	
	local texthide = findfeature("text","is","hide")
	
	if (texthide == nil) then
		for i,rules in ipairs(visualfeatures) do
			local text = ""
			local rule = rules[1]
			
			for a,b in ipairs(rule) do
				text = text .. b .. " "
			end
			
			local conds = rules[2]
			if (#conds > 0) then
				for a,cond in ipairs(conds) do
					text = text .. cond[1] .. " "
					
					if (cond[2] ~= nil) then
						if (#cond[2] > 0) then
							for c,d in ipairs(cond[2]) do
								text = text .. d .. " "
							end
						end
					end
				end
			end
			
			MF_write(text,i)
		end
	end
end

function mapcells()
	local count = 0 
	
	for i,v in pairs(unitmap) do
		if (#v > 0) then
			count = count + 1
		end
	end
	
	return count
end

function findfeatureat(rule1,rule2,rule3,x,y)
	local result = {}
	local targets = findfeature(rule1,rule2,rule3)
	
	if (targets ~= nil) then
		for i,v in ipairs(targets) do
			local name = v[1]
			for a,unit in ipairs(units) do
				local unitx,unity = unit.values[XPOS],unit.values[YPOS]
				if (unit.values[XPOS] == x) and (unit.values[YPOS] == y) then
					if (unit.strings[UNITNAME] == name) or ((unit.strings[UNITTYPE] == "text") and (name == "text")) then
						local conds = v[2]
						if testcond(conds,unit.fixed) then
							table.insert(result, unit.fixed)
						end
					end
				end
			end
		end
	end
	
	if (#result > 0) then
		return result
	else
		return nil
	end
end

function copy(unitid,x,y)
	for i,unit in ipairs(units) do
		if (unit.fixed == unitid) then
			local oldx,oldy,dir,name,float = unit.values[XPOS],unit.values[YPOS],unit.values[DIR],unit.strings[UNITNAME],unit.values[FLOAT]
			
			local this = create(name,x,y,dir,oldx,oldy,float)
			return this
		end
	end
end

function create(name,x,y,dir,oldx_,oldy_,float_)
	local oldx,oldy,float = x,y,0
	local tileid = x + y * roomsize
	
	if (oldx_ ~= nil) then
		oldx = oldx_
	end
	
	if (oldy_ ~= nil) then
		oldy = oldy_
	end
	
	if (float_ ~= nil) then
		float = float_
	end
	
	local newunitid = MF_emptycreate(name,oldx,oldy)
	local newunit = mmf.newObject(newunitid)
	
	local id = newid()
	
	newunit.values[ONLINE] = 1
	newunit.values[XPOS] = x
	newunit.values[YPOS] = y
	newunit.values[DIR] = dir
	newunit.values[ID] = id
	newunit.values[FLOAT] = float
	
	newunit.flags[9] = true
	addunit(newunitid)
	addunitmap(newunitid,x,y)
	dynamic(newunitid)
	
	addundo({"create",name,id})
	
	return newunit.fixed
end

function getunitid(id)
	for i,unit in ipairs(units) do
		if (unit.values[ID] == id) then
			return unit.fixed
		end
	end
	
	return 0
end

function newid()
	local result = currid
	currid = currid + 1
	return result
end

function addunitmap(id,x,y)
	local doadd = true
	local tileid = x + y * roomsize
	
	if (unitmap[tileid] == nil) then
		unitmap[tileid] = {}
	else
		for a,b in ipairs(unitmap[tileid]) do
			if (b == id) then
				doadd = false
			end
		end
	end
	
	if doadd then
		table.insert(unitmap[tileid], id)
	end
end

function updateunitmap(id,oldx,oldy,x,y)
	local tileid = x + y * roomsize
	local oldtileid = oldx + oldy * roomsize
	
	if (unitmap[oldtileid] ~= nil) then
		for i,v in ipairs(unitmap[oldtileid]) do
			if (v == id) then
				table.remove(unitmap[oldtileid], i)
			end
		end
	end
	
	addunitmap(id,x,y)
end

function inside(name,x,y,dir,unitid)
	local ins = {}
	if (featureindex[name] ~= nil) then
		for i,rule in ipairs(featureindex[name]) do
			local baserule = rule[1]
			local conds = rule[2]
			
			local target = baserule[1]
			local verb = baserule[2]
			local object = baserule[3]
			
			if (target == name) and (verb == "has") then				
				table.insert(ins, {object,conds})
			end
		end
	end
	
	if (#ins > 0) then
		for i,v in ipairs(ins) do
			local object = v[1]
			local conds = v[2]
			if testcond(conds,unitid) then
				if (object ~= "text") then
					for a,mat in pairs(objectlist) do
						if (a == object) and (object ~= "empty") and (object ~= "group") then
							if (object ~= "all") then
								create(object,x,y,dir)
							else
								createall(v,x,y,unitid)
							end
						end
					end
				else
					create("text_" .. name,x,y,dir)
				end
			end
		end
	end
end

function hasfeature(rule1,rule2,rule3,unitid)
	if (featureindex[rule1] ~= nil) and (rule2 ~= nil) and (rule3 ~= nil) then
		for i,rules in ipairs(featureindex[rule1]) do
			local rule = rules[1]
			local conds = rules[2]
			
			if (conds[1] ~= "never") then
				if (rule[1] == rule1) and (rule[2] == rule2) and (rule[3] == rule3) then
					if testcond(conds,unitid) then
						return true
					end
				end
			end
		end
	end
	
	if (featureindex[rule3] ~= nil) and (rule2 ~= nil) and (rule1 ~= nil) then
		for i,rules in ipairs(featureindex[rule3]) do
			local rule = rules[1]
			local conds = rules[2]
			
			if (conds[1] ~= "never") then
				if (rule[1] == rule1) and (rule[2] == rule2) and (rule[3] == rule3) then
					if testcond(conds,unitid) then
						return true
					end
				end
			end
		end
	end
	
	if (featureindex[rule2] ~= nil) and (rule1 ~= nil) and (rule3 == nil) then
		local usable = false
		
		if (featureindex[rule1] ~= nil) then
			for i,rules in ipairs(featureindex[rule1]) do
				local rule = rules[1]
				local conds = rules[2]
				
				if (conds[1] ~= "never") then
					for a,mat in pairs(objectlist) do
						if (a == rule[1]) then
							usable = true
						end
					end
					
					if (rule[1] == rule1) and (rule[2] == rule2) and usable then
						if testcond(conds,unitid) then
							return true
						end
					end
				end
			end
		end
	end
	
	return nil
end

function animate()
	for i,unitid in ipairs(animunits) do
		local unit = mmf.newObject(unitid)
		local name = unit.strings[UNITNAME]
		local sleep = hasfeature(name,"is","sleep",unitid)
		
		if (unit.values[TILING] == 3) then
			if (sleep == nil) then
				unit.values[VISUALDIR] = (unit.values[VISUALDIR] + 1) % 2
				unit.direction = ((unit.values[DIR] * 8 + unit.values[VISUALDIR]) + 32) % 32
			else
				unit.values[VISUALDIR] = -1
				unit.direction = ((unit.values[DIR] * 8 + unit.values[VISUALDIR]) + 32) % 32
			end
		end
		
		if (unit.values[TILING] == 2) then
			if (sleep == nil) then
				unit.direction = ((unit.values[DIR] * 8 + unit.values[VISUALDIR]) + 32) % 32
			else
				unit.values[VISUALDIR] = -1
				unit.direction = ((unit.values[DIR] * 8 + unit.values[VISUALDIR]) + 32) % 32
			end
		end
	end
end

function issolid(unitid)
	local unit = mmf.newObject(unitid)
	local name = unit.strings[UNITNAME]
	
	if (unit.strings[UNITTYPE] == "text") then
		name = "text"
	end
	
	local ispush = hasfeature(name,"is","push",unitid)
	local ispull = hasfeature(name,"is","pull",unitid)
	local ismove = hasfeature(name,"is","move",unitid)
	local isyou = hasfeature(name,"is","you",unitid)
	
	if (ispush ~= nil) or (ispull ~= nil) or (ismove ~= nil) or (isyou ~= nil) then
		return true
	end
	
	return false
end

function isgone(unitid)
	if (issafe(unitid) == false) then
		local unit = mmf.newObject(unitid)
		local x,y,name = unit.values[XPOS],unit.values[YPOS],unit.strings[UNITNAME]
		
		if (unit.strings[UNITTYPE] == "text") then
			name = "text"
		end
		
		local isyou = hasfeature(name,"is","you",unitid)
		local ismelt = hasfeature(name,"is","melt",unitid)
		local isweak = hasfeature(name,"is","weak",unitid)
		local isshut = hasfeature(name,"is","shut",unitid)
		local isopen = hasfeature(name,"is","open",unitid)
		local ismove = hasfeature(name,"is","move",unitid)
		local ispush = hasfeature(name,"is","push",unitid)
		local ispull = hasfeature(name,"is","pull",unitid)
		local eat = findfeatureat(nil,"eat",name,x,y)
		
		if (eat ~= nil) then
			for i,v in ipairs(eat) do
				if (v ~= unitid) then
					return true
				end
			end
		end

		local issink = findfeatureat(nil,"is","sink",x,y)
		
		if (issink ~= nil) then
			for i,v in ipairs(issink) do
				if (v ~= unitid) then
					return true
				end
			end
		end
		
		if (isyou ~= nil) then
			local ispoof = findfeatureat(nil,"is","poof",x,y)
			
			if (ispoof ~= nil) then
				return true
			end
		end
		
		if (ismelt ~= nil) then
			local ishot = findfeatureat(nil,"is","hot",x,y)
			
			if (ishot ~= nil) then
				return true
			end
		end
		
		if (isshut ~= nil) then
			local isopen_ = findfeatureat(nil,"is","open",x,y)
			
			if (isopen_ ~= nil) then
				return true
			end
		end
		
		if (isopen ~= nil) then
			local isshut_ = findfeatureat(nil,"is","shut",x,y)
			
			if (isshut_ ~= nil) then
				return true
			end
		end
		
		if (isweak ~= nil) then
			local things = findallhere(x,y)
			
			if (#things > 1) then
				return true
			end
		end
	end
	
	return false
end

function floating(id1,id2)
	local unit1 = mmf.newObject(id1)
	local unit2 = mmf.newObject(id2)
	
	if (unit1.values[FLOAT] == unit2.values[FLOAT]) then
		return true
	end
	
	return false
end

function findgroup()
	local result = {}
	local groupstuff = {}
	
	if (featureindex["group"] ~= nil) then
		for i,rules in ipairs(featureindex["group"]) do
			local rule = rules[1]
			local conds = rules[2]

			if (rule[3] == "group") then
				if (rule[2] == "is") and (rule[1] ~= "group") then
					table.insert(groupstuff, {rule[1],conds})
				end
			end
		end
	end
	
	if (#groupstuff > 0) then
		for i,v in ipairs(groupstuff) do
			for a,mat in pairs(objectlist) do
				if (a ~= nil) then
					if (a == v[1]) and (a ~= "group") then
						table.insert(result, v)
					end
				end
			end
		end
	end
	
	return result
end

function issafe(unitid)
	local unit = mmf.newObject(unitid)
	local name = unit.strings[UNITNAME]
	
	if (unit.strings[UNITTYPE] == "text") then
		name = "text"
	end
	
	local safe = hasfeature(name,"is","safe",unitid)
	
	if (safe ~= nil) then
		return true
	end
	
	return false
end

function issleep(unitid)
	local unit = mmf.newObject(unitid)
	local name = unit.strings[UNITNAME]
	
	if (unit.strings[UNITTYPE] == "text") then
		name = "text"
	end
	
	local sleep = hasfeature(name,"is","sleep",unitid)
	
	if (sleep ~= nil) then
		return true
	end
	
	return false
end

function testcond(conds,unitid,x_,y_)
	local result = true
	
	local x,y,name = 0,0,""
	if (unitid > 2) then
		local unit = mmf.newObject(unitid)
		x = unit.values[XPOS]
		y = unit.values[YPOS]
		name = unit.strings[UNITNAME]
	elseif (unitid == 2) then
		x = x_
		y = y_
		name = "empty"
	end
	
	if (unitid == 0) then
		print("WARNING!! Unitid is zero!!")
	end
	
	if (conds ~= nil) then
		if (#conds > 0) then
			local valid = false
			
			for i,cond in ipairs(conds) do
				local condtype = cond[1]
				local params = cond[2]
				
				if (condtype == "never") then
					result = false
					valid = true
				elseif (condtype == "on") then
					valid = true
					local allfound = 0
					local alreadyfound = {}
					
					local tileid = x + y * roomsize
					
					if (#params > 0) then
						for a,b in ipairs(params) do
							if (b ~= "empty") then
								if (unitmap[tileid] ~= nil) then
									for c,d in ipairs(unitmap[tileid]) do
										if (d ~= unitid) then
											local unit = mmf.newObject(d)
											local name_ = unit.strings[UNITNAME]
											
											if (unit.strings[UNITTYPE] == "text") then
												name_ = "text"
											end
											
											if (name_ == b) and (alreadyfound[b] == nil) then
												alreadyfound[b] = 1
												allfound = allfound + 1
											end
										end
									end
								else
									print("unitmap is nil at " .. tostring(x) .. ", " .. tostring(y) .. " for object " .. unit.strings[UNITNAME] .. " (" .. tostring(unitid) .. ")!")
								end
							else
								result = false
								--[[
								local onempty = true
								
								if (unitmap[tileid] ~= nil) then
									if (#unitmap[tileid] > 1) then
										onempty = false
									end
								end
								
								if onempty then
									alreadyfound[b] = 1
									allfound = allfound + 1
								end
								]]--
							end	
						end
					else
						print("no parameters given!")
					end

					if (allfound ~= #params) then
						result = false
					end
				elseif (condtype == "near") then
					valid = true
					local allfound = 0
					local alreadyfound = {}
					
					if (#params > 0) then
						for a,b in ipairs(params) do
							for g=-1,1 do
								for h=-1,1 do
									if (b ~= "empty") then
										local tileid = (x + g) + (y + h) * roomsize
										if (unitmap[tileid] ~= nil) then
											for c,d in ipairs(unitmap[tileid]) do
												if (d ~= unitid) then
													local unit = mmf.newObject(d)
													local name_ = unit.strings[UNITNAME]
													
													if (unit.strings[UNITTYPE] == "text") then
														name_ = "text"
													end
													
													if (name_ == b) and (alreadyfound[b] == nil) then
														alreadyfound[b] = 1
														allfound = allfound + 1
													end
												end
											end
										end
									else
										local nearempty = true
								
										local tileid = (x + g) + (y + h) * roomsize
										if (unitmap[tileid] ~= nil) then 
											if (#unitmap[tileid] > 1) then
												nearempty = false
											end
										end
										
										if nearempty then
											alreadyfound[b] = 1
											allfound = allfound + 1
										end
									end
								end
							end
						end
					else
						print("no parameters given!")
					end

					if (allfound ~= #params) then
						result = false
					end
				elseif (condtype == "not on") then
					valid = true
					local tileid = x + y * roomsize
					
					if (#params > 0) then
						for a,b in ipairs(params) do
							if (b ~= "empty") then
								if (unitmap[tileid] ~= nil) then
									for c,d in ipairs(unitmap[tileid]) do
										if (d ~= unitid) then
											local unit = mmf.newObject(d)
											local name_ = unit.strings[UNITNAME]
											
											if (unit.strings[UNITTYPE] == "text") then
												name_ = "text"
											end
											
											if (name_ == b) then
												result = false
											end
										end
									end
								else
									print("unitmap is nil at " .. tostring(x) .. ", " .. tostring(y) .. "!")
								end
							else
								local onempty = true

								if (unitmap[tileid] ~= nil) then 
									if (#unitmap[tileid] > 1) then
										onempty = false
									end
								end
								
								if onempty then
									result = false
								end
							end
						end
					else
						print("no parameters given!")
					end
				elseif (condtype == "not near") then
					valid = true
					
					if (#params > 0) then
						for a,b in ipairs(params) do
							for g=-1,1 do
								for h=-1,1 do
									if (b ~= "empty") then
										local tileid = (x + g) + (y + h) * roomsize
										if (unitmap[tileid] ~= nil) then
											for c,d in ipairs(unitmap[tileid]) do
												local unit = mmf.newObject(d)
												local name_ = unit.strings[UNITNAME]
												
												if (unit.strings[UNITTYPE] == "text") then
													name_ = "text"
												end
												
												if (name_ == b) and (alreadyfound[b] == nil) then
													result = false
												end
											end
										end
									else
										local nearempty = true
								
										local tileid = (x + g) + (y + h) * roomsize
										if (unitmap[tileid] ~= nil) then 
											if (#unitmap[tileid] > 1) then
												nearempty = false
											end
										end
										
										if nearempty then
											result = false
										end
									end
								end
							end
						end
					else
						print("no parameters given!")
					end
				elseif (condtype == "oldalone") then
					valid = true
					
					for g=-1,1 do
						for h=-1,1 do
							local tileid = (x + g) + (y + h) * roomsize
							if (unitmap[tileid] ~= nil) then
								for c,d in ipairs(unitmap[tileid]) do
									if (d ~= unitid) then
										result = false
									end
								end
							end
						end
					end
				elseif (condtype == "not oldalone") then
					valid = true
					local alone = true
					
					for g=-1,1 do
						for h=-1,1 do
							local tileid = (x + g) + (y + h) * roomsize
							if (unitmap[tileid] ~= nil) then
								for c,d in ipairs(unitmap[tileid]) do
									if (d ~= unitid) then
										alone = false
									end
								end
							end
						end
					end

					if (alone == true) then
						result = false
					end
				elseif (condtype == "alone") then
					valid = true
					
					local tileid = x + y * roomsize
					if (unitmap[tileid] ~= nil) then
						for c,d in ipairs(unitmap[tileid]) do
							if (d ~= unitid) then
								result = false
							end
						end
					end
				elseif (condtype == "not alone") then
					valid = true
					
					local tileid = x + y * roomsize
					if (unitmap[tileid] ~= nil) then
						if (#unitmap[tileid] == 1) then
							result = false
						end
					end
				end
			end
			
			if (valid == false) then
				print("zero conditions!")
				result = true
			end
		end
	end
	
	return result
end

function restore()
	clearunits()
	local allunits = MF_getunits()
	
	if (#allunits > 0) then
		for i,v in ipairs(allunits) do
			addunit(v)
		end
	end
	
	updatecode = 1
	code()
end

function comparerules(baserule1,baserule2)
	local same = true
	
	for i,v in ipairs(baserule1) do
		if (v ~= baserule2[i]) then
			same = false
		end
	end
	
	return same
end

function debugfunction()
	if ruledebug then
		ruledebug = false
		print("Rule debugging disabled")
	else
		ruledebug = true
		print("Rule debugging enabled")
	end
	
	for a,b in pairs(objectlist) do
		print(a)
	end
	
	restore()
end

function getmat(m)
	local found = false
	
	for i,v in pairs(objectlist) do
		if (i == m) then
			found = true
		end
	end
	
	if found then
		return m
	else
		return nil
	end
end

function destroylevel()
	local dellist = {}
	for i,unit in ipairs(units) do
		table.insert(dellist, unit.fixed)
	end
	
	if (#dellist > 0) then
		for i,unitid in ipairs(dellist) do
			local unit = mmf.newObject(unitid)
			local c1,c2 = getcolour(unitid)
			MF_particles("bling",unit.values[XPOS],unit.values[YPOS],10,c1,c2)
			delete(unitid,0)
		end
	end
end

function addaction(unitid,data)
	local unit = nil
	
	if (unitid ~= nil) and (unitid > 0) then
		unit = mmf.newObject(unitid)
	end
	
	local invalid = false
	
	if (unit ~= nil) then
		invalid = unit.flags[15]
	end
	
	local action = {}
	
	if (invalid == false) then
		table.insert(action, unitid)
		
		for i,v in ipairs(data) do
			table.insert(action, v)
		end
		
		table.insert(updatelist, action)
	end
end

function doupdate()
	if (#updatelist > 0) then
		for i,data in ipairs(updatelist) do
			--MF_alert("Updating unit " .. tostring(data[1]) .. " to coordinates " .. tostring(data[2]) .. ", " .. tostring(data[3]) .. " with direction " .. tostring(data[4]))
			
			if (data[2] == "update") then
				if (data[3] ~= nil) and (data[4] ~= nil) then
					update(data[1],data[3],data[4],data[5])
				else
					print("No meaningful update data for object " .. tostring(data[1]) .. ", " .. data[2])
				end
			elseif (data[2] == "updatedir") then
				if (data[3] ~= nil) then
					updatedir(data[1],data[3])
				else
					print("No meaningful update data for object " .. tostring(data[1]) .. ", " .. data[2])
				end
			elseif (data[2] == "convert") or (data[2] == "emptyconvert") then
				if (data[3] ~= nil) then
					doconvert(data)
				else
					print("No meaningful update data for object " .. tostring(data[1]) .. ", " .. data[2])
				end
			elseif (data[2] == "createall") then
				if (data[3] ~= nil) then
					createall(data[3])
				else
					print("No meaningful update data for object " .. tostring(data[1]) .. ", " .. data[2])
				end
			end
		end
	end
	
	updatelist = {}
end